@@ -1,13 +1,15 @@
 """MESSENGER UVVS data class"""
 import numpy as np
+from scipy.optimize import minimize_scalar
 import pandas as pd
-import bokeh.plotting as plt
-from bokeh.models import HoverTool
+import bokeh.plotting as bkp
+from bokeh.models import HoverTool, Whisker
 from bokeh.palettes import Set1
+from bokeh.io import export_png
 from astropy import units as u
 from astropy.visualization import PercentileInterval
 from .database_setup import database_connect
-from nexoclom import Input, LOSResult
+from nexoclom import Input
 
 
 class InputError(Exception):
@@ -177,8 +179,7 @@ class MESSENGERdata:
         self.data = None
         self.taa = None
         self.inputs = None
-        self.model_strength = None
-        self.model_label = None
+        self.model_info = None
         
         if species is None:
             pass
@@ -202,9 +203,9 @@ class MESSENGERdata:
             try:
                 with database_connect() as con:
                     data = pd.read_sql(query, con)
-            except:
-                raise Input('MESSENGERdata.__init__',
-                            'Problem with comparisons given.')
+            except Exception:
+                raise InputError('MESSENGERdata.__init__',
+                                 'Problem with comparisons given.')
 
             if len(data) > 0:
                 self.species = species
@@ -236,7 +237,7 @@ class MESSENGERdata:
     def __len__(self):
         try:
             return len(self.data)
-        except:
+        except Exception:
             return 0
 
     def __getitem__(self, q_):
@@ -255,8 +256,7 @@ class MESSENGERdata:
         new.query = self.query
         new.taa = self.taa
         new.data = self.data.iloc[q].copy()
-        new.model_strength = self.model_strength
-        new.model_label = self.model_label
+        new.model_info = self.model_info
         new.inputs = self.inputs
 
         return new
@@ -308,7 +308,44 @@ class MESSENGERdata:
             assert 0, 'You somehow picked a bad combination.'
 
     def model(self, inputs_, npackets, quantity='radiance',
-              dphi=3*u.deg, overwrite=False, filenames=None, label=None):
+              fit_method='chisqmin', dphi=3*u.deg, overwrite=False,
+              masking=None, filenames=None, label=None):
+        """Run the nexoclom model with specified inputs and fit to the data.
+        
+        ** Parameters**
+        inputs
+            A nexoclom Input object or a string with the path to an inputs file.
+            
+        npackets
+            Number of packets to run.
+            
+        fit_method
+            Allows user to specify the quantity to be minimized when fitting
+            the model to the data. Default = chisqmin (chi-squared
+            minimization). Other option is 'difference' which minimizes the
+            sum of the differences of the data and the model ignoring the
+            data uncertainty.
+            
+        dphi
+            Angular size of the view cone. Default = 3 deg. Must be
+            given as an astropy quantity.
+            
+        overwrite
+            Set to True to erase any previous model runs with these inputs.
+            Default = False
+        
+        masking
+            Allows user to specify which data points are included in the fit
+            to the model. Default = None
+            * middleX: Use the middle X% of values. For example, middle50
+              excludes the faintest and brightest 25% spectra
+            
+            * minsnrX: specifies the minimum signal-to-noise ratio to use.
+              minsnr3 exludes any spectra with SNR < 3
+            
+            * siglimitX: Excludes any data where the model is not within X σ
+              of the data in an initial fit to the data.
+        """
 
         if isinstance(inputs_, str):
             inputs = Input(inputs_)
@@ -319,12 +356,19 @@ class MESSENGERdata:
 
         # TAA needs to match the data
         oldtaa = inputs.geometry.taa
-        if len(set(self.data.orbit.tolist())) == 1:
-            inputs.geometry.taa = np.median(self.data.taa)*u.rad
-        elif np.max(self.data.taa)-np.min(self.data.taa) < 3*np.pi/180.:
+        if len(self.data.orbit.unique()) == 1:
             inputs.geometry.taa = np.median(self.data.taa)*u.rad
+        elif self.data.taa.max()-self.data.taa.min() < 3*np.pi/180.:
+            inputs.geometry.taa = self.data.taa.median()*u.rad
         else:
             assert 0, 'Too wide a range of taa'
+            
+        # If using a planet-fixed source map, need to set subsolarlon
+        if ((inputs.spatialdist.type == 'surface map') and
+            (inputs.spatialdist.coordinate_system == 'planet-fixed')):
+            inputs.spatialdist.subsolarlon = self.data.subslong.median()*u.rad
+        else:
+            pass
 
         # Run the model
         inputs.run(npackets, overwrite=overwrite)
@@ -342,39 +386,90 @@ class MESSENGERdata:
         
         # modkey is the number for this model
         modkey = f'model{len(self.inputs)-1:00d}'
+        packkey = f'packets{len(self.inputs)-1:00d}'
+        maskkey = f'mask{len(self.inputs)-1:00d}'
         self.data[modkey] = model_result.radiance/1e3 # Convert to kR
-
-        # Estimate model strength (source rate) by mean of middle 50%
-        interval = PercentileInterval(50)
-        lim = interval.get_limits(self.data.radiance)
-        mask = ((self.data.radiance >= lim[0]) &
-                (self.data.radiance <= lim[1]))
-
-        strunit = u.def_unit('10**26 atoms/s', 1e26/u.s)
-        m_data = np.mean(self.data.radiance[mask])
-        m_model = np.mean(self.data[modkey][mask])
-        model_strength = m_data/m_model * strunit * strunit
-        self.data[modkey] = self.data[modkey] * model_strength.value
-
-        if self.model_strength is None:
-            self.model_strength = {modkey: model_strength}
-        else:
-            self.model_strength[modkey] = model_strength
+        self.data[packkey] = model_result.packets
+        
+        strength, goodness_of_fit, mask = self.fit_model(modkey, fit_method, masking)
+        self.data[modkey] = self.data[modkey]*strength.value
+        self.data[maskkey] = mask
 
         if label is None:
             label = modkey.capitalize()
         else:
             pass
-        
-        if self.model_label is None:
-            self.model_label = {modkey: label}
-        else:
-            self.model_label[modkey] = label
 
-        print(f'Model strength for {label} = {model_strength}')
+        model_info = {'fit_method': fit_method,
+                      'goodness-of-fit': goodness_of_fit,
+                      'strength': strength,
+                      'label': label}
+        if self.model_info is None:
+            self.model_info = {modkey: model_info}
+        else:
+            self.model_info[modkey] = model_info
+        
+        print(f'Model strength for {label} = {strength}')
 
         # Put the old TAA back in.
         inputs.geometry.taa = oldtaa
+    
+    def fit_model(self, modkey, fit_method, masking):
+        def chisq(x):
+            return np.sum((self.data[mask]['radiance'] -
+                           x * self.data[mask][modkey])**2 /
+                          self.data[mask]['sigma']**2)/(sum(mask) - 1)
+
+        def difference(x):
+            return np.sum(np.abs(self.data[mask]['radiance'] -
+                                 x * self.data[mask][modkey]))
+
+        mask = np.array([True for _ in self.data.radiance])
+        sigmalimit = None
+        if masking is not None:
+            for masktype in masking.split(';'):
+                masktype = masktype.strip().lower()
+                if masktype.startswith('middle'):
+                    perinterval = float(masktype[6:])
+                    # Estimate model strength (source rate) by fitting middle %
+                    interval = PercentileInterval(perinterval)
+                    lim = interval.get_limits(self.data.radiance)
+                    mask = (mask &
+                            (self.data.radiance >= lim[0]) &
+                            (self.data.radiance <= lim[1]))
+                elif masktype.startswith('minalt'):
+                    minalt = float(masktype[6:])
+                    mask = mask & (self.data.alttan >= minalt)
+                elif masktype.startswith('minsnr'):
+                    minSNR = float(masktype[6:])
+                    snr = self.data.radiance / self.data.sigma
+                    mask = mask & (snr > minSNR)
+                elif masktype.startswith('siglimit'):
+                    sigmalimit = masktype
+                else:
+                    raise InputError('MESSENGERdata.fit_model',
+                                     f'masking = {masktype} not defined.')
+        else:
+            pass
+        
+        strunit = u.def_unit('10**26 atoms/s', 1e26/u.s)
+        available_fitfunctions = ['chisqmin', 'diffmin']
+        if fit_method.lower() in available_fitfunctions:
+            model_strength = minimize_scalar(fit_method.lower())
+            if sigmalimit is not None:
+                siglimit = float(sigmalimit[8:])
+                diff = ((self.data['radiance']-
+                         model_strength.x*self.data[modkey]) /
+                        self.data['sigma'])
+                mask = mask & (diff < siglimit*self.data['sigma'])
+                model_strength = minimize_scalar(fit_method.lower())
+            else:
+                pass
+        else:
+            raise InputError('MESSENGERdata.fit_model',
+                             f'fit_method = {fit_method} not defined.')
+        
+        return model_strength.x * strunit, model_strength.fun, mask
 
     def plot(self, filename=None, show=True, **kwargs):
         if filename is not None:
@@ -382,7 +477,7 @@ class MESSENGERdata:
                 filename += '.html'
             else:
                 pass
-            plt.output_file(filename)
+            bkp.output_file(filename)
         else:
             pass
 
@@ -391,10 +486,12 @@ class MESSENGERdata:
             lambda x: x.isoformat()[0:19])
 
         # Put the dataframe in a useable form
-        source = plt.ColumnDataSource(self.data)
+        self.data['lower'] = self.data.radiance - self.data.sigma
+        self.data['upper'] = self.data.radiance + self.data.sigma
+        source = bkp.ColumnDataSource(self.data)
 
         # Make the figure
-        fig = plt.figure(plot_width=1200, plot_height=800,
+        fig = bkp.figure(plot_width=1200, plot_height=800,
                          x_axis_type='datetime',
                          title=f'{self.species}, {self.query}',
                          x_axis_label='UTC',
@@ -403,34 +500,40 @@ class MESSENGERdata:
 
         # plot the data
         dplot = fig.circle(x='utc', y='radiance', size=7, color='black',
-                           legend='Data', hover_color='white', source=source)
+                           legend_label='Data', hover_color='white',
+                           source=source)
+        
+        # Add error bars
+        fig.add_layout(Whisker(source=source, base='utc', upper='upper',
+                               lower='lower'))
 
         # tool tips
-        tips = [('index', '$index'), ('UTC', '@utcstr'),
+        tips = [('index', '$index'),
+                ('UTC', '@utcstr'),
                 ('Radiance', '@radiance{0.2f} kR')]
-        if  self.model_label is not None:
-            for modkey, modlabel in self.model_label.items():
-                tips.append((modlabel, f'@{modkey}{{0.2f}} kR'))
+        if self.model_info is not None:
+            for modkey, info in self.model_info.items():
+                tips.append((info['label'], f'@{info["strength"]}{{0.2f}} kR'))
         datahover = HoverTool(tooltips=tips,
                               renderers=[dplot])
         fig.add_tools(datahover)
 
         # Plot the model
         col = (c for c in Set1[9])
-        if self.model_label is not None:
-            for modkey, modlabel in self.model_label.items():
+        if self.model_info is not None:
+            for modkey, info in self.model_info.items():
                 try:
                     c = next(col)
                 except StopIteration:
                     col = (c for c in Set1[9])
                     c = next(col)
 
-                f = fig.line(x='utc', y=modkey, source=source,
-                               legend=modlabel, color=c)
-                f = fig.circle(x='utc', y=modkey, size=7, source=source,
-                               legend=modlabel, color=c)
-                datahover.renderers.append(f)
-                print(modlabel)
+                f0 = fig.line(x='utc', y=modkey, source=source,
+                               legend_label=info['label'], color=c)
+                f1 = fig.circle(x='utc', y=modkey, size=7, source=source,
+                               legend_label=info['label'], color=c)
+                datahover.renderers.append(f0)
+                datahover.renderers.append(f1)
 
         # Labels, etc.
         fig.title.align = 'center'
@@ -441,15 +544,19 @@ class MESSENGERdata:
         fig.legend.click_policy = 'hide'
 
         if filename is not None:
-            plt.output_file(filename)
-            plt.save(fig)
+            bkp.output_file(filename)
+            export_png(fig, filename=filename.replace('.html', '.png'))
+            bkp.save(fig)
+            
         else:
             pass
 
         if show:
-            plt.show(fig)
+            bkp.show(fig)
 
-    def export(self, filename, columns=['utc', 'radiance']):
+        return fig
+
+    def export(self, filename, columns=('utc', 'radiance')):
         """Export data and models to a file.
         **Parameters**
         
@@ -469,9 +576,9 @@ class MESSENGERdata:
         No outputs.
         
         """
-        columns_ = columns.copy()
-        if self.model_label is not None:
-            columns_.extend(self.model_label.keys())
+        columns_ = list(columns)
+        if self.model_info is not None:
+            columns_.extend(self.model_info.keys())
         else:
             pass
         
